"""Class defintitions for Media types.

`Media` objects represent a file to be managed and its metadata. The `Media` class
itself is a base class from which subtypes should be derived. It provides an interface
and some template methods for its subtypes, but is generally not very useful on its own.

`Media.library`: A `pathlib.Path` that represents the root location a `Media` object
should be moved to when `Media.organize()` is called.

`Media.path`: A `pathlib.Path` that represents the actual location of the file on disk.
This value is optional, and falsey values should set it to `None`.

`Media.infer()`: A method that attempts to infer metadata about the file (e.g. from its
filename) and updates the appropriate metadata attributes with this information.

`Media.organize()`: A method that attempts to move the file to a new location, updates
the `path` attribute on success, and returns this new value.

`Media.metadata`: A property that returns all metadata as a dictionary, intended for
ease of inspection and discovery.
"""

import logging
import pathlib
import re
from dataclasses import dataclass, field
from shutil import chown, move
from string import capwords
from typing import Any, Pattern

from nielsen.config import config

logger: logging.Logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


@dataclass(slots=True)
class Media:
    """Media objects represent a file to be managed and its metadata."""

    patterns: list[Pattern] = field(
        default_factory=list,
        init=False,
        repr=False,
        hash=False,
        compare=False,
        metadata=None,
    )
    _path: pathlib.Path = field(
        hash=True,
    )
    _section: str = field(
        init=False,
        repr=False,
        hash=False,
        compare=False,
        metadata=None,
    )
    _library: pathlib.Path = field(
        init=False,
        repr=False,
        hash=False,
        compare=False,
        metadata=None,
    )
    _metadata: dict[str, Any] = field(
        default_factory=dict,
        init=False,
        repr=True,
        hash=True,
        compare=True,
        metadata=None,
    )

    def __post_init__(self):
        """Modify instance after initialization."""

        # Set the path again to run it through the property setter
        self.path = self.path

        # By default, instances of a subclass should use the section of the config
        # corresponding to their type name.
        self.section = self.__class__.__name__.lower()

        # Similarly, the library should be pulled from the section of the config
        # corresponding to the type name.
        self.library = config.getpath(self.section, "library").resolve()  # type: ignore

        # Load filename patterns for the type.
        self.load_patterns()

    @property
    def path(self) -> pathlib.Path:
        """Return a Path object representing the Media file on disk."""

        return self._path

    @path.setter
    def path(self, value: pathlib.Path | str) -> None:
        """Set the path property. Attempt to coerce the value to a Path if not
        provided as such."""

        try:
            if not isinstance(value, pathlib.Path):
                value = pathlib.Path(value)

                # TODO Do not allow non-files
                if not value.is_file():
                    raise TypeError(repr(value))
        except TypeError:
            logger.exception("Media.path must be a file: %s", repr(value))
            raise

        self._path = value.resolve()

    @property
    def library(self) -> pathlib.Path:
        """Return a Path object representing the root of the library for the Media type
        from the config."""

        return self._library

    @library.setter
    def library(self, value: pathlib.Path | str) -> None:
        """Set the library property. Attempt to coerce the value to a Path if not
        provided as such."""

        try:
            if not isinstance(value, pathlib.Path):
                value = pathlib.Path(value)

            if not value.is_dir():
                raise TypeError(repr(value))
        except TypeError:
            logger.exception("Media.library must be a directory: %s", repr(value))
            raise

        self._library = value.resolve()

    def load_patterns(self) -> None:
        """Load filename patterns for the instance type into the patterns property."""

        self._load_patterns()
        logger.debug("Loaded %s patterns.", self.__class__.__name__)

    def _load_patterns(self) -> None:
        """Should only be called directly by the `load_patterns` method."""

    @property
    def section(self) -> str:
        """Return a string denoting which section of the ConfigParser this Media should
        refer to when looking for its options."""

        return self._section

    @section.setter
    def section(self, value: str) -> None:
        """Set the section in the ConfigParser to refer to for options."""

        if not config.has_section(value):
            config.add_section(value)

        self._section = value

    @property
    def orgdir(self) -> pathlib.Path:
        """Return the orgdir property."""

        raise NotImplementedError

    def infer(self) -> None:
        """Infer information about the object's metadata based on its filename. Does
        some basic error checking and then sets the various metadata fields using the
        `metadata.setter` method which each subclass must implement to handle value
        transformations and/or formatting."""

        if not self.patterns:
            logger.error("NO_PATTERNS: No patterns defined to match against.")
            logger.debug(repr(self))
            return

        metadata: dict[str, Any] = self._match()
        self.metadata = metadata

    def _match(self) -> dict[str, Any]:
        """Return the results of `Match.groupdict()` if the filename matched any of the
        patterns. Should only be called by `infer()`."""

        # Assert values here which have already been checked by Media.infer to avoid
        # type errors from the linter.
        assert self.path

        for pattern in self.patterns:
            match = pattern.fullmatch(self.path.name)
            if match:
                return match.groupdict()

        logger.info("NO_MATCH: %s did not match any filename patterns.", self.path.name)
        return {}

    def organize(self) -> pathlib.Path:
        """Move the file to the appropriate media library on disk, set file ownership
        and mode."""

        if not self.path.is_file():
            logger.error(
                "Path attribute does not point to a regular file: %s", self.path
            )
            raise TypeError(self.path)

        if not self.library.is_dir():
            try:
                self.library.mkdir(exist_ok=True)
            except (PermissionError, NotADirectoryError):
                logger.exception(
                    "CANNOT_ORGANIZE: Library directory does not exist and could not be created: %s",
                    self.library,
                )
                raise

        # Ensure the orgdir exists and move the file there.
        logger.info("Move %s → %s/.", self.path.name, self.orgdir)
        if not config.getboolean("nielsen", "simulate"):
            self.orgdir.mkdir(exist_ok=True, parents=True)
            self.path = pathlib.Path(
                move(self.path, self.orgdir / self.path.name)
            ).resolve()
            logger.debug("New path: %s", self.path)

        self.path.chmod(int(config.get(self.section, "mode"), 8))

        # If chown gets a None value for user or group, it won't modify the existing
        # values. Use None as a fallback to leave things alone unless the user has
        # explicitly set a different value in their configuration.
        user: str | None = config.get(self.section, "owner", fallback=None)
        group: str | None = config.get(self.section, "group", fallback=None)

        if user or group:
            chown(self.path, user, group)  # type: ignore

        return self.path

    @property
    def metadata(self) -> dict[str, Any]:
        """Return the metadata property, a dictionary of information relevant to the
        Media type."""

        raise NotImplementedError

    @metadata.setter
    def metadata(self, metadata: dict[str, Any]) -> None:
        """Set the metadata property. Must be a dictionary, but subclasses should
        implement their own transformations (if any) by implementing this setter
        method."""

        raise NotImplementedError

    # This implementation may not be especially useful for base Media objects, but it
    # provides a sensible default implementation other classes are free to use,
    # extend, or override.
    def rename(self) -> pathlib.Path:
        """Rename the file associated with this Media object to match its `str`
        representation, but do not change its parent directory. Return the new location
        as a Path object."""

        if not self.path or not self.path.exists():
            raise FileNotFoundError(self.path)

        simulate: bool = config.getboolean("nielsen", "simulate")
        dest: pathlib.Path = self.path.with_stem(f"{self!s}")
        logger.info("Renaming %s → %s. Simulate: %s", self.path, dest, simulate)

        # TODO: Raise an exception file destination conflicts
        # TODO: Handle exception on the client side
        if dest.exists():
            if self.path.samefile(dest):
                logger.info("File already named correctly.")
            else:
                logger.warning("FILE_CONFLICT: %s already exists.", dest)

            return self.path

        if not simulate:
            self.path = self.path.rename(dest)

        return self.path

    def transform(self, field: str) -> str:
        """Transform a field's value based on the corresponding config section. For
        example, passing `field=series` for a `TV` object will look for an option in the
        `tv/transform/series` section of the config matching `self.series`. If found,
        `self.series` will be replaced with the option value. If not found, nothing
        changes."""

        section: str = f"{self.section}/transform/{field}"
        option: str = getattr(self, field)

        # Create the corresponding transform sub-section if it doesn't exist.
        if not config.has_section(section):
            logger.warning("NO_TRANSFORM_SECTION: %s", section)
            return option

        if not config.has_option(section, option):
            logger.warning("NO_TRANSFORM_OPTION: %s", option)
            return option

        transformed: str = config.get(section, option)
        logger.info("TRANSFORM: %s → %s", option, transformed)

        return transformed

    def __str__(self) -> str:
        """Return a friendly, human-readable version of the file path, fit for
        renaming or display purposes."""

        return f"{self.path.resolve()!s}"


@dataclass(order=True, slots=True)
class TV(Media):
    """A Media subclass for TV shows."""

    series: str = ""
    series_id: int = 0
    season: int = 0
    episode: int = 0
    title: str = ""

    def _load_patterns(self) -> None:
        """Load filename patterns for the instance type into the patterns property."""

        self.patterns = [
            # The.Glades.S02E01.Family.Matters.HDTV.XviD-FQM.avi
            re.compile(
                r"(?P<series>.+?)\.+(?P<year>\d{4}|\(\d{4}\))?\.*S(?P<season>\d{2})\.?E(?P<episode>\d{2})\.*(?P<title>.*)?\.+(?P<extension>\w+)$",
                re.IGNORECASE,
            ),
            # The.Flash.2014.217.Flash.Back.HDTV.x264-LOL[ettv].mp4
            re.compile(
                r"(?P<series>.+?)\.+(?P<year>\d{4}|\(\d{4}\))?\.(?P<season>\d{1,2})(?P<episode>\d{2})\.*(?P<title>.*)?\.+(?P<extension>\w+)$",
                re.IGNORECASE,
            ),
            # the.glades.201.family.matters.hdtv.xvid-fqm.avi
            re.compile(
                r"(?P<series>.+)\.+S?(?P<season>\d{1,})\.?E?(?P<episode>\d{2,})\.*(?P<title>.*)?\.+(?P<extension>\w+)$",
                re.IGNORECASE,
            ),
            # The Glades -02.01- Family Matters.avi
            re.compile(
                r"(?P<series>.+)\s+-(?P<season>\d{2})\.(?P<episode>\d{2})-\s*(?P<title>.*)\.(?P<extension>.+)$"
            ),
            # The Glades -201- Family Matters.avi
            re.compile(
                r"(?P<series>.+[^\s-])[\s-]+(?P<season>\d{1,2})(?P<episode>\d{2,})[\s-]+(?P<title>.*)\.(?P<extension>.+)$"
            ),
            # Last ditch effort to get essential information
            re.compile(
                r"(?P<series>.+)S(?P<season>\d{1,2})E(?P<episode>\d{2,})(?P<title>.*)\.(?P<extension>.+)$"
            ),
        ]

    @property
    def metadata(self) -> dict[str, Any]:
        """Return a dictionary of metadata relevant to the TV type."""

        return {
            "series": self.series,
            "season": self.season,
            "episode": self.episode,
            "title": self.title,
        }

    @metadata.setter
    def metadata(self, metadata: dict[str, Any]) -> None:
        """Transform values from the given metadata dictionary and use it to set the
        object's fields."""

        self.series = metadata.get("series", "").replace(".", " ").strip().title()
        self.series = self.transform("series")
        self.season = int(metadata.get("season", 0))
        self.episode = int(metadata.get("episode", 0))
        self.title = metadata.get("title", "").replace(".", " ").strip()

        tags: re.Pattern = re.compile(
            r"\(?(1080p|720p|HDTV|WEB|PROPER|REPACK|RERIP)\)?.*", re.IGNORECASE
        )
        # Use string.capwords() rather than str.title() to properly handle letters after apostrophes.
        self.title = capwords(re.sub(tags, "", self.title).strip())

    @property
    def orgdir(self) -> pathlib.Path:
        """Return the orgdir property."""

        return pathlib.Path(self.library / f"{self.series}/Season {self.season:02d}/")

    def __str__(self) -> str:
        """Return a friendly, human-readable version of the file metadata, fit for
        renaming or display purposes."""

        return f"{self.series or 'Unknown'} -{self.season:02d}.{self.episode:02d}- {self.title or 'Unknown'}"

    def __repr__(self) -> str:
        """Return a string with enough information to recreate the object."""

        return f"<{self.__class__.__name__}({self.path=}, {self.series=}, {self.season=}, {self.episode=}, {self.title=})>"


# vim: tabstop=4 softtabstop=4 shiftwidth=4 expandtab textwidth=88



"""A collection of classes used to orchestrate combinations of Media and Fetcher
operations to be used by any frontend application.

Rather than creating individual Media instances of a specific subclass and a Fetcher
instance and manually calling all of their individual methods, Processors provide a
convenience method (process) which creates a Media instance of the desired type and
handles all changes to that Media instance."""

from dataclasses import dataclass
from enum import Enum
import pathlib

from nielsen.config import config
import nielsen.fetcher
import nielsen.media


class MediaType(str, Enum):
    TV = "tv"


@dataclass
class Processor:
    """A convenience class to handle all the Media and Fetcher operations required to
    manage an individual file. The `media_type` is a Media class and is used to
    construct instances of the given type. The `fetcher` is an instance of an object
    implementing the `Fetcher` Protocol."""

    media_type: type[nielsen.media.Media]
    fetcher: nielsen.fetcher.Fetcher

    def process(self, path: pathlib.Path) -> nielsen.media.Media:
        """Convenience function to process Media for storage in its library, respecting the configuration
        options (so some steps may be skipped).
        1. Infer metadata based on the Media type.
        2. Fetch missing data.
        3. Rename the file.
        4. Move the file to the appropriate directory.
        5. Modify file ownership and mode.
        """

        # Create an instance of the appropriate Media subclass
        media: nielsen.media.Media = self.media_type(path)

        media.infer()

        if config.getboolean(media.section, "fetch"):
            self.fetcher.fetch(media)

        if config.getboolean(media.section, "rename"):
            media.rename()

        if config.getboolean(media.section, "organize"):
            media.organize()

        return media


@dataclass
class ProcessorFactory:
    """A Factory class for Processors. Given a Media and Fetcher class, calling an
    instance of this Factory will return a Processor with the given Media class and
    Fetcher instance."""

    media_type: type[nielsen.media.Media]
    fetcher: type[nielsen.fetcher.Fetcher]

    def __call__(self) -> Processor:
        """Return a Processor for the given types."""

        return Processor(self.media_type, self.fetcher())


PROCESSOR_FACTORIES: dict[MediaType, ProcessorFactory] = {
    MediaType.TV: ProcessorFactory(nielsen.media.TV, nielsen.fetcher.TVMaze),
}



#!/usr/bin/env python3

"""Generate pickled objects for Mocks to return in unit tests. This allows us to work
with real data, but prevents us from having to make additional calls to an external API
every time the tests are run."""

import logging
import pathlib
import pickle
from typing import Any

import requests

logger: logging.Logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.StreamHandler())


def main():
    """Pickle responses from external API calls for each fetcher."""

    tvmaze()


def pickle_data(data: Any, path: str) -> None:
    """Pickle some arbitrary data and store it in the fixtures directory under the path
    described by the given string."""

    # Ensure all parent directories are created for the file. The fixtures directory
    # should be a sibling of the `bin` directory which contains this script.
    file: pathlib.Path = (
        pathlib.Path(__file__).parent.parent.parent / "fixtures/" / pathlib.Path(path)
    )
    file.parent.mkdir(parents=True, exist_ok=True)
    file.write_bytes(pickle.dumps(data))


def tvmaze() -> None:
    """Make calls to the TVMaze API using the `requests` module, then pickle the
    responses for later loading and use by the unit tests for the TVMaze fetcher."""

    TVMAZE: str = "https://api.tvmaze.com"
    TRANSLATION = str.maketrans({" ": "-", "=": "-", "&": "-"})
    queries: list[dict[str, str]] = [
        {"api": "search/shows", "params": "q=Agents+of+SHIELD"},
        {"api": "search/shows", "params": "q=Ted+Lasso"},
        {"api": "search/shows", "params": "q=useless+search+string"},
        {"api": "seasons/112939/episodes", "params": ""},
        {"api": "shows/44458/episodebynumber", "params": "season=1&number=3"},
        {"api": "shows/44458/seasons", "params": ""},
        {"api": "singlesearch/shows", "params": "q=Ted+Lasso"},
        {"api": "singlesearch/shows", "params": "q=useless+search+string"},
    ]

    for query in queries:
        stem: str = f"{query['api']}-{query['params'].lower()}".translate(TRANSLATION)
        path: str = f"tv/{stem}.pickle".translate(TRANSLATION)
        uri: str = f"{TVMAZE}/{query['api']}?{query['params']}"
        response: requests.models.Response = requests.get(uri)

        logger.debug(response.json())

        pickle_data(response, path)


if __name__ == "__main__":
    main()


# vim: tabstop=4 softtabstop=4 shiftwidth=4 expandtab textwidth=88



#!/usr/bin/env python3
"""Subcommands for managing TV shows with Nielsen."""

import logging
from pathlib import Path
from typing import Any, Optional

import typer
from requests import Response
from rich.pretty import pprint
from typing_extensions import Annotated

import nielsen.config
import nielsen.fetcher
import nielsen.media
from nielsen.config import config as config

logger: logging.Logger = logging.getLogger(__name__)
config_files: list[str] = nielsen.config.load_config()

app: typer.Typer = typer.Typer(name="TV")


@app.command(no_args_is_help=True)
def fetch(
    series: Annotated[
        Optional[str], typer.Option(help="Show name to search for")
    ] = None,
    series_id: Annotated[
        Optional[int], typer.Option(help="The TVMaze series ID")
    ] = None,
    season: Annotated[
        Optional[int],
        typer.Option(help="Season number (to get additional details)"),
    ] = None,
    episode: Annotated[
        Optional[int],
        typer.Option(help="Episode number (to get additional details)"),
    ] = None,
    interactive: Annotated[
        bool,
        typer.Option(
            is_flag=True,
            help="Interactively select from multiple results",
        ),
    ] = True,
    raw: Annotated[
        bool,
        typer.Option(is_flag=True, help="Print the JSON formatted API response"),
    ] = False,
) -> None:
    """Fetch information about a TV show. The more information provided to the command,
    the more specific the returned results."""

    if series and series_id:
        typer.echo("The --series and --series-id options are mutually exclusive.")
        raise typer.Exit(1)
    elif not (series or series_id):
        typer.echo("One of --series or --series-id required.")
        raise typer.Exit(2)

    if episode and not season:
        typer.echo("Cannot fetch episode information without a season.")
        raise typer.Exit(3)

    fetcher: nielsen.fetcher.TVMaze = nielsen.fetcher.TVMaze()

    if series:
        series_id = fetcher.get_series_id(series, interactive)

    if series_id and season and episode:
        response: Response = fetcher.episodebynumber(series_id, season, episode)
        data: Any = response.json()

        if raw:
            pprint(data)
        else:
            typer.echo(fetcher.pretty_episode(data))

    elif series_id and season:
        season_id: int = fetcher.get_season_id(series_id, season)
        response = fetcher.seasons_episodes(season_id)
        data = response.json()

        if raw:
            pprint(data)
        else:
            typer.echo(fetcher.pretty_season(data))

    elif series_id:
        response = fetcher.shows(series_id)
        data = response.json()

        if raw:
            pprint(data)
        else:
            typer.echo(fetcher.pretty_series(data))

    else:
        logger.critical("Could not fetch any information.")
        raise typer.Exit(4)

    if response.ok:
        nielsen.config.update_config(Path(config_files[-1]))


@app.command()
def apply(
    files: Annotated[list[str], typer.Argument(help="File(s) to process")],
    series: Annotated[Optional[str], typer.Option(help="Show name")],
    season: Annotated[
        int,
        typer.Option(help="Season number (to get additional details)"),
    ],
    episode: Annotated[
        Optional[int],
        typer.Option(help="Episode number (to get additional details)"),
    ] = None,
    interactive: Annotated[
        bool,
        typer.Option(
            is_flag=True,
            help="Interactively select from multiple results",
        ),
    ] = True,
    simulate: Annotated[
        bool,
        typer.Option(help="Show file operations without performing them"),
    ] = config.getboolean("nielsen", "simulate"),
) -> None:
    """Apply episode information to the provided file(s)."""

    if not series:
        typer.echo("--series is required.")
        raise typer.Exit(1)

    if episode and not season:
        typer.echo("Cannot fetch episode information without a season.")
        raise typer.Exit(2)

    config.set("nielsen", "simulate", str(simulate))

    fetcher: nielsen.fetcher.TVMaze = nielsen.fetcher.TVMaze()

    series_id = fetcher.get_series_id(series, interactive)

    if series_id and season and episode:
        if len(files) != 1:
            typer.echo("Must specify exactly one file to apply this data to.")
            raise typer.Exit(3)

        # Create a TV instance and attach metadata to it for the Fetcher to work with
        media: nielsen.media.TV = nielsen.media.TV(
            Path(files.pop()), series=series, season=season, episode=episode
        )
        fetcher.fetch(media)
        media.rename()

        return

    season_id: int = fetcher.get_season_id(series_id, season)
    response: Response = fetcher.seasons_episodes(season_id)
    episodes: list[dict[str, Any]] = response.json()

    for e, f in zip(episodes, files):
        media = nielsen.media.TV(Path(f), series=series, season=season)
        typer.echo(
            f"Apply information from {series} Season {season} Episode {e['number']} to {media.path}"
        )
        media.episode = e.get("number", 0)
        media.title = e.get("name", "")
        media.rename()


if __name__ == "__main__":
    app()



"""Fetchers are used to query remote sources for metadata about a given Media object."""

import logging
import urllib.parse
from html.parser import HTMLParser
from io import StringIO
from typing import Any, Callable, Optional, Protocol

import requests

import nielsen.media
from nielsen.config import config

logger: logging.Logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Fetcher(Protocol):
    """Used to fetch metadata from an external source rather than infering it from the
    file name."""

    def fetch(self, media: nielsen.media.Media) -> None:
        """Fetch and update metadata using information from the given `Media` object."""
        ...


class TVMaze:
    """Fetch metadata for TV shows using the TVMaze API."""

    SERVICE: str = "https://api.tvmaze.com"
    IDS: str = "tvmaze/ids"

    def fetch(self, media: nielsen.media.TV) -> None:
        """Fetch metadata from TVMaze, update the metadata of the provided Media object,
        record the series ID in the config."""

        series_id: Optional[int] = self.get_series_id(
            media.series, config.getboolean("DEFAULT", "interactive")
        )

        if not series_id:
            raise ValueError("No Series ID")

        self.set_series_id(media.series, str(series_id))
        media.title = self.get_episode_title(media)

    def set_series_id(self, series: str, id: int | str) -> None:
        """Create a mapping from a series name to a TVMaze series ID in the config."""
        # TODO: Ensure the config gets written back to disk.

        logger.debug("Set '%s' to TVMaze Series ID %s.", series, id)

        if not config.has_section(self.IDS):
            logger.debug("Adding '%s' section to config.", self.IDS)
            config.add_section(self.IDS)

        config.set(self.IDS, series, str(id))

    def get_series_id(self, series: str, interactive: bool = False) -> int:
        """Return the TVMaze ID for the series. Will check for a local config file first
        and search TVMaze if a local match isn't found. Optionally, prompt the user to
        select the correct series interactively if multiple results are found. Returns 0
        if no series ID can be found."""

        if config.has_option(self.IDS, series):
            lookup = self.get_series_id_local
        elif interactive:
            lookup = self.get_series_id_search
        else:
            lookup = self.get_series_id_singlesearch

        series_id: int = lookup(series)
        logger.debug("Series: %s, ID: %s", series, series_id)

        if series_id and not config.has_option(self.IDS, series):
            self.set_series_id(series, series_id)

        return series_id

    def get_series_id_local(self, series: str) -> int:
        """Get the series ID from the configuration."""

        return config.getint(self.IDS, series, fallback=0)

    def get_series_id_search(
        self,
        series: str,
        picker: Callable[[str, list[dict[Any, Any]]], dict[str, Any]] | None = None,
    ) -> int:
        """Get the series ID from the TVMaze `search` endpoint, which can return
        multiple results. If multiple results are returned, prompt the user to pick one.
        The `picker` defines this behavior.
        """

        response: requests.Response = self.search_shows(series)
        rjson: list[dict[Any, Any]] = response.json()

        series_id: int = 0
        # If TVMaze returns an empty list, return 0.
        if not rjson:
            return series_id

        # If only one result, add it to the config and return the ID.
        if len(rjson) == 1:
            return rjson[0]["show"]["id"]

        # If multiple results, offer a means of selecting the correct one.
        if picker is None:
            picker = self.pick_series

        selection: dict = picker(series, rjson)

        series_id = selection["show"]["id"]

        return series_id

    def get_series_id_singlesearch(self, series: str) -> int:
        """Get the series ID from the TVMaze `singlesearch` endpoint, which returns a
        single result that it considers the best match.
        URL: /singlesearch/shows?q=:query
        """

        response: requests.Response = self.search_shows_single(series)
        rjson: dict[Any, Any] = response.json()

        if response.ok:
            return rjson.get("id", 0)

        return 0

    def get_episode_title(self, media: nielsen.media.TV) -> str:
        """Return the episode title for the given media object from the TVMaze API.
        URL: /shows/:id/episodebynumber?season=:season&number=:number.
        """

        series_id: Optional[int] = self.get_series_id(media.series)
        episode_title: str = ""

        if not series_id:
            raise ValueError("No Series ID")

        if not media.season:
            raise ValueError("No Season")

        if not media.episode:
            raise ValueError("No Episode Number")

        response: requests.Response = self.episodebynumber(
            series_id, media.season, media.episode
        )
        rjson: dict[Any, Any] = response.json()

        if response.ok:
            episode_title = str(rjson.get("name"))

        return episode_title

    def get_season_id(self, series_id: int, season: int) -> int:
        """Return information about a given `season` of a given `series_id`.
        URL: /shows/:id/seasons"""

        response: requests.Response = self.shows_seasons(series_id)
        rjson: dict[Any, Any] = response.json()

        for item in rjson:
            match item:
                case {"number": number} if number == season:
                    return item["id"]

        return 0

    def search_shows(self, series: str) -> requests.Response:
        """Search TVMaze for the given `series` and return the `requests.Response`
        object. URL: /search/shows?q=:query."""

        request: str = (
            f"{self.SERVICE}/search/shows/?q={urllib.parse.quote_plus(series)}"
        )
        logging.debug("Series: %s\nRequest: %r", series, request)
        response: requests.Response = requests.get(request)
        logging.debug(response)

        return response

    def search_shows_single(self, series: str) -> requests.Response:
        """Search TVMaze for the given `series` and return the `requests.Response`
        object containing information about the single best result.
        URL: /singlesearch/shows?q=:query"""

        request: str = (
            f"{self.SERVICE}/singlesearch/shows/?q={urllib.parse.quote_plus(series)}"
        )
        logger.debug("Series: %r\nRequest: %r", series, request)
        response: requests.Response = requests.get(request)
        logger.debug(response)

        return response

    def episodebynumber(
        self, series: int | str, season: int, episode: int
    ) -> requests.Response:
        """URL: /shows/:id/episodebynumber?season=:season&number=:episode"""

        if isinstance(series, int):
            series_id: int = series
        elif isinstance(series, str):
            series_id = self.get_series_id(series)

        if not series_id:
            raise ValueError("No Series ID")

        request: str = f"{self.SERVICE}/shows/{series_id}/episodebynumber?season={season}&number={episode}"
        response: requests.Response = requests.get(request)

        return response

    def seasons_episodes(self, season_id: int) -> requests.Response:
        """URL: /seasons/:id/episodes"""

        request: str = f"{self.SERVICE}/seasons/{season_id}/episodes"
        logger.debug("Request: %s", request)
        response: requests.Response = requests.get(request)

        return response

    def shows(self, series_id: int) -> requests.Response:
        """URL: /shows/:id"""

        request: str = f"{self.SERVICE}/shows/{series_id}"
        logger.debug("Request: %s", request)
        response: requests.Response = requests.get(request)

        return response

    def shows_seasons(self, series_id: int) -> requests.Response:
        """URL: /shows/:id/seasons"""

        request: str = f"{self.SERVICE}/shows/{series_id}/seasons"
        logger.debug("Request: %s", request)
        response: requests.Response = requests.get(request)

        return response

    @staticmethod
    def pick_series(query: str, results: list[dict[Any, Any]]) -> dict[str, Any]:
        """Display a text-based picker to choose a single show from multiple results."""

        print(f"Search results for: {query}")

        for option, result in enumerate(results, start=1):
            match result:
                # Network Television
                case {
                    "show": {
                        "name": name,
                        "id": series_id,
                        "premiered": premiered,
                        "network": {"name": nw_name, "country": nw_country},
                    }
                }:
                    logger.debug("Matched Network")
                    name = name
                    series_id = series_id
                    premiered = premiered
                    network: str = nw_name
                    country: str = (
                        nw_country.get("name", "Unknown") if nw_country else "Unknown"
                    )

                # Streaming Platforms
                case {
                    "show": {
                        "name": name,
                        "id": series_id,
                        "premiered": premiered,
                        "webChannel": {"name": wc_name, "country": wc_country},
                    }
                }:
                    logger.debug("Matched Streaming")
                    name = name
                    series_id = series_id
                    premiered = premiered
                    network = wc_name
                    country = (
                        wc_country.get("name", "Unknown") if wc_country else "Unknown"
                    )

                # Minimal Results
                case {
                    "show": {
                        "name": name,
                        "id": series_id,
                        "premiered": premiered,
                    }
                }:
                    logger.debug("Matched Minimal")
                    name = name
                    series_id = series_id
                    premiered = premiered
                    network = "Unknown"
                    country = "Unknown"

                # This should never happen
                case _:
                    # NOTE: It may be better to simply skip these results.
                    logger.error("Unable to parse search result")
                    logger.debug(result)
                    name = "Unknown"
                    series_id = 0
                    premiered = "Unknown"
                    network = "Unknown"
                    country = "Unknown"

            print(
                f"{option}. {name} (Premiered: {premiered}, Network: {network}, Country: {country}, ID: {series_id})"
            )

        # Start with an invalid selection to start the loop
        selection: int = -1
        while 0 > selection or selection >= len(results):
            selection = int(input("Select series: ")) - 1

        return results[selection]

    @staticmethod
    def pretty_series(data: dict[str, Any]) -> str:
        """Return a nicely formatted string of high-level information about the series
        as a whole, such as the title, TVMaze ID, premiere date, whether it's currently
        airing, and a plot summary.

        This is intended to work with the results of the TVMaze `/shows/:id`
        endpoint."""

        return "\n".join(
            (
                f"{data['name']} - ID: {data['id']} - {data['url']}",
                f"Premiered: {data['premiered']} - Status: {data['status']}",
                f"{strip_tags(data['summary'])}",
            )
        )

    @staticmethod
    def pretty_season(data: list[dict[str, Any]]) -> str:
        """Return a nicely formatted string containing information about every episode
        in a season, including the season and episode numbers, episode title, a link
        to the episode URL on TVMaze, and an episode summary.

        This is intended to work with the results of the TVMaze `/seasons/:id/episodes`
        endpoint."""

        pretty_episodes: list[str] = []

        for episode in data:
            pretty_episodes.append(TVMaze.pretty_episode(episode))

        return "\n\n".join(pretty_episodes)

    @staticmethod
    def pretty_episode(data: dict[str, Any]) -> str:
        """Return a nicely formatted string of episode specific information, including
        the season and episode numbers, episode title, a link to the episode URL on
        TVMaze, and an episode summary.

        This is intended to work with the results of the TVMaze `/seasons/:id/episodes`
        and `/shows/:id/episodebynumber?season=:season&number=:number` endpoints."""

        return "\n".join(
            (
                f"{data['season']}x{data['number']} - {data["name"]}",
                f"{data['url']}",
                f"{strip_tags(data['summary'])}",
            )
        )


class MLStripper(HTMLParser):
    """Markup Language Stripper"""

    def __init__(self):
        super().__init__()
        self.reset()
        self.strict = False
        self.convert_charrefs = True
        self.text = StringIO()

    def handle_data(self, data: str) -> None:
        self.text.write(data)

    def get_data(self) -> str:
        return self.text.getvalue()


def strip_tags(html):
    """Strip HTML tags to make API responses more readable in the terminal."""

    s = MLStripper()
    s.feed(html)
    return s.get_data()


# vim: tabstop=4 softtabstop=4 shiftwidth=4 expandtab textwidth=88
